/*
* Rascan is a rewrite of Inter7's simscan
* 
* Copyright (C) 2006 Roberto Alsina <ralsina@kde.org>
* Copyright (C) 2004-2005 Inter7 Internet Technologies, Inc.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later
* version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software Foundation,
* Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*
*/

#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <stdio.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include "bstrlib.h"

#include "rascan.h"

// Globals
struct timeval start, stop;
int DebugFlag = 0;
bstring workdir;
bstring unique_ext;
bstring message_name;
bstring spam_message_name;
bstring buffer;
bstring addr_name;
bstring addr_buf;
bstring MailFrom;
bstring VirusName;
struct bstrList *addresses;
bstring runned_scanners;
bstring ClamavVersion;
bstring SAVersion;
int FoundVirus;
double utime;
bstring Subject;

float SpamHits;
float ReqHits;
int InHeaders;
int IsSpam;

// Prototypes

int
main (int argc, char **argv)
{
  bstring tmpstr;
  int ret;
  int pid;
  workdir = bfromcstr (getenv ("RASCAN_DIR"));
  if (workdir == NULL)
    {
      workdir = bfromcstr (WORKDIRDEFAULT);
    }
  Subject = bfromcstr ("");
  runned_scanners = bfromcstr ("");
  VirusName = bfromcstr ("");

  // get the starttime of the process
  gettimeofday (&start, (struct timezone *) 0);

  tmpstr = bfromcstr (getenv ("RASCAN_DEBUG"));
  if (tmpstr)
    {
      DebugFlag = atoi (tmpstr->data);
    }

  //FIXME: implement attachment blocking
  //init_attach();

  /* format the new directory name */
  format_dir (workdir);
  debug (bformat ("work dir: %s", workdir->data), 0);

  // create the working directory, allow group access too
  if (mkdir (workdir->data, 0750) == -1)
    {
      debug (bformat
             ("error making work dir, temporary error: %s",
              strerror (errno)), 0);
      _exit (EXIT_400);
    }
  umask (0077);

  // change to the new working directory
  if (chdir (workdir->data) != 0)
    {
      debug (bformat
             ("error changing directory to workdir: %s",
              strerror (errno)), 0);
      exit_clean (EXIT_400);
    }

  // From fd 0 we read the envelope
  // From fd 1 we read the message

  FILE *f0 = fdopen (0, "r");
  if (f0 == NULL)
    {
      debug (bformat
             ("error fdopening fd 0 for msg: %s", strerror (errno)), 0);
      exit_clean (EXIT_400);
    }
  FILE *f1 = fdopen (1, "r");
  if (f1 == NULL)
    {
      debug (bformat
             ("error fdopening fd 1 for envelope %s", strerror (errno)), 0);
      exit_clean (EXIT_400);
    }

  // open a msg file to hold the email
  // and an addr file to hold the addresses

  message_name = bformat ("msg.%s", unique_ext->data);
  addr_name = bformat ("addr.%s", unique_ext->data);

  FILE *fmsg = fopen (message_name->data, "w");
  FILE *faddr = fopen (addr_name->data, "w");

  if (NULL == fmsg)
    {
      debug (bformat ("error opening msg file: %s ", strerror (errno)), 0);
      exit_clean (EXIT_400);
    }
  if (NULL == faddr)
    {
      debug (bformat ("error opening addr name: %s\n", addr_name), 0);
      exit_clean (EXIT_400);
    }


  //write msg from fd0 to msg file

  for (;;)
    {
      bstring line = bgets ((bNgetc) fgetc, f0, '\n');
      if (line == NULL)
        break;
      if (fwrite (line->data, line->slen, 1, fmsg) != 1)
        {
          debug (bformat ("error writing msg: %s", strerror (errno)), 0);
          /* on a write error close the file so we can remove the directory */
          fclose (fmsg);
          exit_clean (EXIT_400);
        }
    }

  //write envelope from fd1 to addr file

  bstring envelope = bfromcstr ("");
  for (;;)
    {
      bstring line = bgets ((bNgetc) fgetc, f1, '\0');
      if (line == NULL)
        {
          break;
        }
      bconcat (envelope, line);
    }

  if (fwrite (envelope->data, envelope->slen, 1, faddr) != 1)
    {
      debug (bformat ("error writing address file: %s", strerror (errno)), 0);
      fclose (faddr);
      exit_clean (EXIT_400);
    }

  // close the addr and msg files

  if (fclose (fmsg) != 0)
    {
      debug (bformat ("error closing msg file: %s", strerror (errno)), 0);
      exit_clean (EXIT_400);
    }
  if (fclose (faddr) != 0)
    {
      debug (bfromcstr ("error closing address file"), 0);
      exit_clean (EXIT_400);
    }

  // Get the mail from value
  addresses = bsplit (envelope, '\0');
  if (addresses == NULL || addresses->qty < 3)
    {
      if (addresses)
        debug (bformat
               ("error: missing pieces of the envelope?, %s %d",
                envelope->data, addresses->qty), 0);
      else
        debug (bformat
               ("error: missing pieces of the envelope?, %s %d",
                envelope->data), 0);
      exit_clean (EXIT_400);
    }

  MailFrom = bmidstr (addresses->entry[0], 1, addresses->entry[0]->slen);

  //FIXME: Implement REGEX stuff
  /*

     if ( check_regex() == 2 ) 
     {
     log_message("REGEX", VirusName, 0);

     reject ("Your email was rejected because it matches a filter");
     }

   */

  /* break the email msg into mime parts */
  if (run_ripmime () != 0)
    {
      debug (bfromcstr ("simscan: ripmime error"), 0);
      exit_clean (EXIT_400);
    }


  //FIXME: implement check_attach
  /* check for attachments to block */
/*  if ( check_attach() > 0 ) 
  {
    log_message("ATTACH", AttachName, 0);
    reject("Your email was rejected because it contains a bad attachment: %s",
       AttachName );
    exit_clean(EXIT_MSG);
  }*/


  lookup_versions ();

  // This is different from simscan, where this is a compile-time
  // option.      
  if (getenv ("DISABLE_CLAMAV") == NULL)
    {

      /* Run ClamAntiVirus, exit on errors */
      ret = check_clam ();
      switch (ret)
        {
        case -1:
          debug (bfromcstr ("fatal error executing clamdscan"), 0);
          exit_clean (EXIT_400);
          break;
        case 1:
          FoundVirus = 1;
          debug (bfromcstr ("clamdscan detected a virus"), 0);
          break;
        default:
          debug (bformat ("normal clamdscan return code: %d", ret), 0);
          break;
        }

      if (FoundVirus == 1)
        {
          debug (bformat
                 ("Your email was rejected because it contains the %s virus",
                  VirusName->data), 0);
          reject (bformat
                  ("Your email was rejected because it contains the %s virus",
                   VirusName->data));
          log_message ("VIRUS DROPPED", VirusName->data, 0);
          /* Drop the message, returning success to sender. */
          exit_clean (EXIT_MSG);
        }
    }

  // This is different from simscan, where this is a compile-time
  // option.      
  if (getenv ("DISABLE_SPAMASSASSIN") == NULL)
    {

      /* optionally check for spam with spamassassin */
      spam_message_name = bformat ("spamc.msg.%s", unique_ext->data);
      ret = check_spam ();

      switch (ret)
        {
          /* spamassassin not enabled for this domain */
        case 2:
          /* re-open the message file file read only */
          /* do nothing, message_name gets openend in any case */
          break;

          /* spam detected */
        case 1:
          debug (bfromcstr ("check_spam detected spam"), 0);
          break;

          /* spamassassin processed message and no spam detected */
        case 0:
          /* open the spam file read only */
          bassign (message_name, spam_message_name);
          break;
          /* errors , return temporary error */
        default:
          debug (bformat ("check_spam had an error ret: %d\n", ret), 0);
          exit_clean (EXIT_400);
        }
    }

  bstring footname = bfromcstr (getenv ("EMAIL_FOOTER"));
  if (footname != NULL)
    {
      debug (bformat ("adding footer %s", footname->data), 0);
      /* Optionally add email footer */
      ret = run_altermime (footname);
      if (ret == -1)
        {
          debug (bformat ("run_altermime had an error ret: %d\n", ret), 0);
          exit_clean (EXIT_400);
        }
    }

  /* re-open the file read only */
  int fd, fd_per;
  if ((fd = open (message_name->data, O_RDONLY)) == -1)
    {
      debug (bformat ("could not re-open file: %s", message_name->data), 0);
      exit_clean (EXIT_400);
    }

  /* re-open the address read only */
  if ((fd_per = open (addr_name->data, O_RDONLY)) == -1)
    {
      debug (bformat ("could not re-open address: %s", addr_name->data), 0);
      exit_clean (EXIT_400);
    }

  /* set the standard input to be the new file */
  if (fd_move (1, fd_per) == -1)
    {
      debug (bfromcstr ("could not fd_move"), 0);
      exit_clean (EXIT_400);
    }

  debug (bfromcstr ("done, execing qmail-queue"), 0);
  int pim[2];
  if (pipe (pim) != 0)
    return (-1);

  /* fork qmail-queue */
  char *qmail_queue;
  qmail_queue = QMAIL_QUEUE;
  switch (pid = vfork ())
    {
    case -1:
      debug (bfromcstr ("error forking qmail-queue"), 0);
      close (pim[0]);
      close (pim[1]);
      exit_clean (EXIT_400);
    case 0:
      close (pim[1]);
      dup2 (pim[0], 0);
      execl (qmail_queue, qmail_queue, 0);
      _exit (-1);
    }
  close (pim[0]);

  gettimeofday (&stop, (struct timezone *) 0);
  utime = SECS (stop) - SECS (start);
  buffer =
    bformat
    ("Received: by simscan %s ppid: %d, pid: %d, t: %3.4fs\n         scanners:%s\n",
     VERSION, getppid (), getpid (), utime, runned_scanners->data);

  if (write (pim[1], buffer->data, buffer->slen) == -1)
    {
      debug (bfromcstr ("error writing received line"), 0);
      exit_clean (EXIT_400);
    }



  /* write the message to qmail-queue */
  while ((ret = read (fd, buffer, sizeof (buffer))) > 0)
    {
      if (write (pim[1], buffer, ret) == -1)
        {
          debug (bformat
                 ("error writing msg to qmail-queue error: %s",
                  strerror (errno)), 0);
          exit_clean (EXIT_400);
        }
    }
  close (pim[1]);
  close (fd);

  /* wait for qmail-queue to finish */
  int qstat;
  if (waitpid (pid, &qstat, 0) == -1)
    {
      debug (bfromcstr ("error forking qmail-queue (back in simscan)"), 0);
      exit_clean (EXIT_400);
    }

  /* hand the email to the qmail-queue */
  debug (bformat ("qmail-queue exited %d\n", WEXITSTATUS (qstat)), 0);

  /* remove the working files */
  remove_files (workdir->data);
  /* pass qmail-queue's exit status on */
  _exit (WEXITSTATUS (qstat));

  /* suppress warning messages */
  return (0);
}

/*
 * format the directory name 
 * use time in seconds . micro seconds . process id
 */
void
format_dir (bstring workdir)
{
  struct timeval mytime;
  gettimeofday (&mytime, (struct timezone *) 0);
  unique_ext =
    bformat ("%ld.%ld.%ld", mytime.tv_sec, mytime.tv_usec,
             (long int) getpid ());
  printf ("unique_ext: %s\n", unique_ext->data);
  bstring tmpstr = bformat ("%s/%s", workdir->data, unique_ext->data);
  bassign (workdir, tmpstr);
  bdestroy (tmpstr);
}


/* Print debug statements according to severity.
 * Threshold is controlled by the RASCAN_DEBUG envvar
 */
void
debug (bstring msg, int priority)
{
  if (DebugFlag > priority)
    {
      fprintf (stderr, "simscan: - %s\n", msg->data);
    }
  bdestroy (msg);
}

void
exit_clean (int error_code)
{
  remove_files (workdir->data);
  debug (bformat ("exit error code: %d\n", error_code), 0);
  _exit (error_code);
}

void
reject (bstring msg)
{
  bstring RejectMsg = bformat ("D%s (%s)", msg->data, VirusName->data);
  write (4, RejectMsg->data, RejectMsg->slen);
  exit_clean (EXIT_MSG);
}

void
log_message (char *state, char *subject, int spam)
{
  int i;
  float reqhits;

  gettimeofday (&stop, (struct timezone *) 0);
  utime = SECS (stop) - SECS (start);


  if (spam == 1)
    {
      reqhits = ReqHits;
      fprintf (stderr, "rascan:[%d]:%s (%.2f/%.2f):%3.4fs:%s:%s:%s:%s",
               getppid (), state, SpamHits, reqhits, utime, subject,
               getenv ("TCPREMOTEIP"), MailFrom->data,
               addresses->entry[1]->data + 1);
    }
  else
    {
      fprintf (stderr, "rascan:[%d]:%s:%3.4fs:%s:%s:%s:%s",
               getppid (), state, utime, subject, getenv ("TCPREMOTEIP"),
               MailFrom->data, addresses->entry[1]->data + 1);
    }

  for (i = 2; i < addresses->qty - 1; i++)
    {
      fprintf (stderr, ",%s", addresses->entry[i]->data + 1);
    }
  fprintf (stderr, "\n");
}

int
run_altermime (bstring footerfile)
{
  bstring cmd = bformat ("altermime --htmltoo --input=%s --disclaimer=%s",
                         message_name->data, footerfile->data);
  debug (bformat ("%s", cmd->data), 0);
  int ret = system (cmd->data);
  /* check if the child died on a signal */
  if (WIFSIGNALED (ret))
    return (-1);
  return (ret);
}


/*
 * break the email into mime parts for scanning 
 */
int
run_ripmime ()
{
  int pid;
  int rmstat;

  /* fork ripmime */
  switch (pid = fork ())
    {
    case -1:
      return (-1);
    case 0:
      close (1);
      close (2);
      execlp ("ripmime", "ripmime", "-i", message_name, "-d", NULL);
      _exit (-1);
    }

  /* wait for ripmime to finish */
  if (waitpid (pid, &rmstat, 0) == -1)
    {
      return (-1);
    }

  /* check if the child died on a signal */
  if (WIFSIGNALED (rmstat))
    return (-1);

  /* if it exited okay, return the status */
  if (WIFEXITED (rmstat))
    {
      return (WEXITSTATUS (rmstat));
    }

  /* should not reach here */
  return (-1);
}


int
check_clam ()                   //FIXME: should use clamdscan's exit code
{
  bstring cmd = bformat ("clamdscan --stdout %s", message_name->data);
  struct tagbstring s1 = bsStatic ("FOUND");
  struct tagbstring s2 = bsStatic (":");

  FILE *pipe = popen (cmd->data, "r");
  int IsClam = 0;
  for (;;)                      //message loop
    {
      bstring line = bgets ((bNgetc) fgetc, pipe, '\n');
      if (line == NULL)
        break;
      if (line->slen < 2)
        continue;
      if (binstr (line, 0, &s1) != BSTR_ERR)    //Virus
        {
          int offset = binstr (line, 0, &s2);
          bassignmidstr (VirusName, line, offset, line->slen - offset - 6);
          IsClam = 1;
          break;                //Report only the first virus
        }
    }
  pclose (pipe);
  add_run_scanner ("clamav", ClamavVersion->data);
  return (IsClam);
}

// Logs the scanner it runs in the msg header

void
add_run_scanner (char *scanner, char *version)
{
  bconcat (runned_scanners, bformat (" %s: %s", scanner, version));
  debug (bformat ("runned_scanners is %s\n", runned_scanners->data), 2);
}

void
lookup_versions ()
{
  FILE *vf = fopen ("/var/qmail/control/rascanversions", "r");

  bstring data = bread ((bNread) fread, vf);
  struct bstrList *versions = bsplit (data, '\n');
  if (versions->qty < 2)
    {
      bstrListDestroy (versions);
      ClamavVersion = bfromcstr ("unknown");
      SAVersion = bfromcstr ("unknown");
      debug (bfromcstr ("Can't open rascanversions, run ragenver"), -1);
    }
  else
    {
      ClamavVersion = versions->entry[0];
      SAVersion = versions->entry[1];
    }
}


int
check_spam ()
{
  int l;
  bstring hits = bfromcstr ("");
  bstring spamc_args = bfromcstr (getenv ("SPAMC_ARGS"));
  if (spamc_args == NULL)
    {
      spamc_args = bfromcstr ("");
    }

  bstring cmd =
    bformat ("spamc %s < %s", spamc_args->data, message_name->data);
  FILE *pipe = popen (cmd->data, "r");
  FILE *fspam = fopen (spam_message_name->data, "w");

  int InHeaders = 1;
  int IsSpam = 0;
  for (;;)                      //message loop
    {
      bstring line = bgets ((bNgetc) fgetc, pipe, '\n');
      if (line == NULL)
        break;

      if (line->slen == 1)
        {
          InHeaders = 0;
        }
      if (InHeaders == 1)       //Parse headers for information
        {
          if (strncmp (line->data, "X-Spam-Flag: YES", 16) == 0)
            {
              IsSpam = 1;
            }

          else if (strncmp (line->data, "Subject:", 8) == 0)
            {
              bassignmidstr (Subject, line, 9, line->slen - 9);
              for (l = 0; Subject->data[l] != 0 && l < sizeof (Subject->data);
                   ++l)
                {
                  if (Subject->data[l] == ':')
                    Subject->data[l] = '_';
                  if (Subject->data[l] == '\r' || Subject->data[l] == '\n')
                    {
                      Subject->data[l] = 0;
                      break;
                    }
                }
            }

          else if (strncmp (line->data, "X-Spam-Status:", 14) == 0)
            {
              int d = 6;
              int offset = binstr (line, 0, bfromcstr ("score="));
              if (offset == BSTR_ERR)
                {
                  offset = binstr (line, 0, bfromcstr ("hits="));
                  d = 5;
                }
              if (offset != BSTR_ERR)
                {
                  offset = offset + d;
                  int offset2 = binstr (line, offset, bfromcstr (" "));
                  if (offset2 == BSTR_ERR)
                    {
                      offset2 = line->slen - offset;
                    }
                  bassignmidstr (hits, line, offset, offset2 - offset);
                  SpamHits = atof (hits->data);
                }

              offset = binstr (line, 0, bfromcstr ("required="));
              if (offset != BSTR_ERR)
                {
                  offset += 9;
                  int offset2 = binstr (line, offset, bfromcstr (" "));
                  if (offset2 == BSTR_ERR)
                    {
                      offset2 = line->slen - offset;
                    }
                  bassignmidstr (hits, line, offset, offset2 - offset);
                }
            }                   //X-Spam-Status
        }                       //InHeaders
      if (fwrite (line->data, line->slen, 1, fspam) != 1)
        {
          debug (bformat ("error writing spam file: %s", strerror (errno)),
                 0);
          return (-1);
        }
    }                           // message loop
  pclose (pipe);
  fclose (fspam);
  add_run_scanner ("spamc", SAVersion->data);
  if (IsSpam == 1)
    {
      log_message ("SPAM", Subject->data, 1);
    }
  else
    {
      log_message ("CLEAN", Subject->data, 1);
    }
  return (0);
}


/*
 * move a file descriptor
 */
int
fd_move (int to, int from)
{
  if (to == from)
    return 0;
  if (fd_copy (to, from) == -1)
    return -1;
  close (from);
  return 0;
}

/*
 * copy a file descriptor
 */
int
fd_copy (int to, int from)
{
  if (to == from)
    return 0;
  if (fcntl (from, F_GETFL, 0) == -1)
    return -1;
  close (to);
  if (fcntl (from, F_DUPFD, to) == -1)
    return -1;
  return 0;
}


/* 
 * From vpopmail source 
 * recursively remove a directory and all it's files
 */
int
remove_files (char *dir)
{
  debug (bformat ("removing workdir %s", dir), 0);
  DIR *mydir;
  struct dirent *mydirent;
  struct stat statbuf;

  /* check the directory stat */
  if (lstat (dir, &statbuf) == 0)
    {

      /* if dir is not a directory unlink it */
      if (!(S_ISDIR (statbuf.st_mode)))
        {
          if (unlink (dir) == 0)
            {
              /* return success we deleted the file */
              return (0);
            }
          else
            {
              /* error, return error to calling function,
               * we couldn't unlink the file
               */
              return (-1);
            }
        }

    }
  else
    {
      /* error, return error to calling function,
       * we couldn't lstat the file
       */
      return (-1);
    }

  /* go to the directory, and check for error */
  if (chdir (dir) == -1)
    {
      /* error, return error to calling function */
      return (-1);
    }

  /* open the directory and check for an error */
  if ((mydir = opendir (".")) == NULL)
    {
      /* error, return error */
      return (-1);
    }

  while ((mydirent = readdir (mydir)) != NULL)
    {

      /* skip the current directory and the parent directory entries */
      if (strcmp (mydirent->d_name, ".") != 0 &&
          strcmp (mydirent->d_name, "..") != 0)
        {

          /* stat the file to check it's type, I/O expensive */
          stat (mydirent->d_name, &statbuf);

          /* Is the entry a directory? */
          if (S_ISDIR (statbuf.st_mode))
            {
              /* delete the sub tree, -1 means an error */
              if (remove_files (mydirent->d_name) == -1)
                {

                  /* on error, close the directory stream */
                  closedir (mydir);

                  /* and return error */
                  return (-1);
                }

              /* the entry is not a directory, unlink it to delete */
            }
          else
            {

              /* unlink the file and check for error */
              if (unlink (mydirent->d_name) == -1)
                {
                  return (-1);
                }
            }
        }
    }

  /* close the directory stream, we don't need it anymore */
  closedir (mydir);
  /* go back to the parent directory and check for error */
  if (chdir ("..") == -1)
    {
      return (-1);
    }

  /* delete the directory, I/O expensive */
  rmdir (dir);

  /* return success */
  return (0);
}
